@@ -20,7 +20,7 @@
 #
 # Contributor(s):
 #  Mikeal Rogers <mikeal.rogers@gmail.com>
-#  Henrik Skupin <hskupin@mozilla.com>
+#  Henrik Skupin <mail@hskupin.info>
 #  Clint Talbert <ctalbert@mozilla.com>
 #
 # Alternatively, the contents of this file may be used under the terms of
@@ -38,16 +38,15 @@
 # ***** END LICENSE BLOCK *****
 
 import copy
-from datetime import datetime, timedelta
-import httplib
 import imp
 import os
 import socket
 import sys
-import threading
 import traceback
-import urllib
-import urlparse
+import urllib2
+
+from datetime import datetime, timedelta
+import manifestparser
 
 try:
     import json
@@ -70,24 +69,12 @@ extension_path = os.path.join(basedir, 'extension')
 
 mozmillModuleJs = "Components.utils.import('resource://mozmill/modules/mozmill.js')"
 
-class ZombieDetector(object):
-    """ Determines if the browser has stopped talking to us.  We assume that
-        if this happens the browser is in a hung state and the test run
-        should be terminated. """
-    # XXX currently unused ;
-    # see https://bugzilla.mozilla.org/show_bug.cgi?id=504440
-
-    def __init__(self, stopfunction):
-        self.stopfunction = stopfunction
-        self.doomsdayTimer = threading.Timer(1800, stopfunction) # 30 minutes
-        self.doomsdayTimer.start()
-
-    def resetTimer(self):
-        # Reset that timer
-        self.doomsdayTimer.cancel()
-        self.doomsdayTimer = threading.Timer(1800, self.stopfunction)
-        self.doomsdayTimer.start()
-
+try:
+    import pkg_resources
+    version = pkg_resources.get_distribution('mozmill').version
+except:
+    # pkg_resources not available
+    version = None
 
 class LoggerListener(object):
     cases = {
@@ -138,7 +125,7 @@ class MozMill(object):
     def __init__(self,
                  runner_class=mozrunner.FirefoxRunner, 
                  profile_class=mozrunner.FirefoxProfile,
-                 jsbridge_port=24242,
+                 jsbridge_port=None,
                  jsbridge_timeout=60):
         """
         - runner_class : which mozrunner class to use
@@ -149,7 +136,7 @@ class MozMill(object):
         
         self.runner_class = runner_class
         self.profile_class = profile_class
-        self.jsbridge_port = jsbridge_port
+        self.jsbridge_port = jsbridge_port or jsbridge.find_port()
         self.jsbridge_timeout = jsbridge_timeout
 
         self.passes = [] ; self.fails = [] ; self.skipped = []
@@ -160,8 +147,7 @@ class MozMill(object):
         self.shutdownModes = enum('default', 'user_shutdown', 'user_restart')
         self.currentShutdownMode = self.shutdownModes.default
         self.userShutdownEnabled = False
-        self.test = None
-        #self.zombieDetector = ZombieDetector(self.stop)
+        self.tests = []
 
         # test time
         self.starttime = self.endtime = None
@@ -175,6 +161,9 @@ class MozMill(object):
         self.add_listener(self.startTest_listener, eventType='mozmill.setTest')
         self.add_listener(self.userShutdown_listener, eventType='mozmill.userShutdown')
 
+        # disable the crashreporter
+        os.environ['MOZ_CRASHREPORTER_NO_REPORT'] = '1'
+
     def add_listener(self, callback, **kwargs):
         self.listeners.append((callback, kwargs,))
 
@@ -224,16 +213,15 @@ class MozMill(object):
 
     def start(self, profile=None, runner=None):
 
-        # Reset our Zombie counter
-        #self.zombieDetector.resetTimer()
-
         if not profile:
             profile = self.profile_class(addons=[jsbridge.extension_path, extension_path])
         self.profile = profile
-        
+
         if not runner:
-            runner = self.runner_class(profile=self.profile, 
-                                       cmdargs=["-jsbridge", str(self.jsbridge_port)])
+            runner = self.runner_class(profile=self.profile)
+
+        if not '-jsbridge' in runner.cmdargs:
+            runner.cmdargs += ['-jsbridge', '%s' % self.jsbridge_port]
 
         self.add_listener(self.firePythonCallback_listener, eventType='mozmill.firePythonCallback')
         self.runner = runner
@@ -247,17 +235,39 @@ class MozMill(object):
         # XXX assumes run_tests will be called soon after (currently true)
         self.starttime = datetime.utcnow()
 
-    def run_tests(self, test, sleeptime=4):
+    def find_tests(self, tests, files=None):
+        if files is None:
+            files = []
+        for test in tests:
+
+            # tests have to be absolute paths to be loaded from JS
+            test = os.path.abspath(test)
+            
+            if os.path.isdir(test):
+                directory = test
+                for f in os.listdir(directory):
+                    if not f.startswith('test'):
+                        continue
+                    path = os.path.join(directory, f)
+                    if os.path.isdir(path):
+                        self.find_tests([path], files)
+                    else:
+                        if f.endswith('.js') and path not in files:
+                            files.append(path)
+            else:
+                files.append(test)
+        return files
+
+
+    def run_tests(self, tests, sleeptime=0):
         """
-        run a test file or directory
-        - test : test file or directory to run
+        run test files or directories
+        - test : test files or directories to run
         - sleeptime : initial time to sleep [s] (not sure why the default is 4)
         """
 
-        # Reset our Zombie Because we are still active
-        #self.zombieDetector.resetTimer()
-
-        self.test = test
+        tests = self.find_tests(tests)
+        self.tests.extend(tests)
 
         frame = jsbridge.JSObject(self.bridge,
                                   "Components.utils.import('resource://mozmill/modules/frame.js')")
@@ -266,10 +276,8 @@ class MozMill(object):
         # transfer persisted data
         frame.persisted = self.persisted
 
-        # run the test directory or file
-        if os.path.isdir(test):
-            frame.runTestDirectory(test)
-        else:
+        # run the test files
+        for test in tests:
             frame.runTestFile(test)
 
         # Give a second for any callbacks to finish.
@@ -280,8 +288,9 @@ class MozMill(object):
         print "TEST-START | %s | %s" % (test['filename'], test['name'])
 
     def endTest_listener(self, test):
-        # Reset our Zombie Counter because we are still active
-        #self.zombieDetector.resetTimer()
+        fname = os.path.split(test['filename'])[1]
+        if fname:
+            test['name'] = "%s::%s" % (fname, test['name'])
 
         self.alltests.append(test)
         if test.get('skipped', False):
@@ -361,7 +370,7 @@ class MozMill(object):
             system = os.environ.get("OS", system).replace('_', ' ')
             service_pack = os.sys.getwindowsversion()[4]
         elif system == "Linux":
-            (distro, version, codename) = platform.linux_distribution()
+            (distro, version, codename) = platform.dist()
             version = distro + " " + version
             if not processor:
                 processor = machine
@@ -395,18 +404,18 @@ class MozMill(object):
 
     def get_report(self):
         """get the report results"""
-        format = "%Y-%m-%dT%H:%M:%S"
+        format = "%Y-%m-%dT%H:%M:%SZ"
 
-        assert self.test, 'run_tests not called'
+        assert self.tests, 'no tests have been run!'
         assert self.starttime, 'starttime not set; have you started the tests?'
         if not self.endtime:
             self.endtime = datetime.utcnow()
 
         report = {'report_type': self.report_type,
+                  'mozmill_version': version,
                   'time_start': self.starttime.strftime(format),
                   'time_end': self.endtime.strftime(format),
                   'time_upload': 'n/a',
-                  'root_path': self.test,
                   'tests_passed': len(self.passes),
                   'tests_failed': len(self.fails),
                   'tests_skipped': len(self.skipped),
@@ -441,29 +450,24 @@ class MozMill(object):
         try:
             # Set the upload time of the report
             now = datetime.utcnow()
-            results['time_upload'] = now.strftime("%Y-%m-%dT%H:%M:%S")
+            results['time_upload'] = now.strftime("%Y-%m-%dT%H:%M:%SZ")
+
+            # Send a POST request to the DB; the POST is implied by the body data
+            body = json.dumps(results)
+            request = urllib2.Request(report_url, body, {"Content-Type": "application/json"})
 
-            # Parse URL fragments and send data
-            url_fragments = urlparse.urlparse(report_url)
-            connection = httplib.HTTPConnection(url_fragments.netloc)
-            connection.request("POST", url_fragments.path, json.dumps(results),
-                               {"Content-type": "application/json"})
-        
             # Get response which contains the id of the new document
-            response = connection.getresponse()
+            response = urllib2.urlopen(request)
             data = json.loads(response.read())
-            connection.close()
-
-            # Check if the report has been created
-            if not data['ok']:
-                print "Creating report document failed (%s)" % data
-                return data
 
             # Print document location to the console and return
             print "Report document created at '%s%s'" % (report_url, data['id'])
             return data
-        except Exception, e:
-            print "Sending results to '%s' failed (%s)." % (report_url, e)
+        except urllib2.HTTPError, e:
+            data = json.loads(e.read())
+            print "Sending results to '%s' failed (%s)." % (report_url, data['reason'])
+        except urllib2.URLError, e:
+            print "Sending results to '%s' failed (%s)." % (report_url, e.reason)
 
     def report(self, report_url):
         """print statistics and send the JSON report"""
@@ -482,7 +486,10 @@ class MozMill(object):
             mozmill.cleanQuit()
         except (socket.error, JSBridgeDisconnectError):
             pass
-
+        except:
+            self.runner.cleanup()
+            raise
+        
         if not close_bridge:
             starttime = datetime.utcnow()
             self.runner.wait(timeout=timeout)
@@ -495,8 +502,7 @@ class MozMill(object):
                 self.runner.wait()
         else: # TODO: unify this logic with the above better
             if hard:
-                self.runner.kill()
-                self.runner.profile.cleanup()
+                self.runner.cleanup()
                 return
 
             # XXX this call won't actually finish in the specified timeout time
@@ -512,8 +518,7 @@ class MozMill(object):
                 x += 1
             else:
                 print "WARNING | endRunner was never called. There must have been a failure in the framework."
-                self.runner.kill()
-                self.runner.profile.cleanup()
+                self.runner.cleanup()
                 sys.exit(1)
 
     def stop(self, fatal=False):
@@ -524,10 +529,8 @@ class MozMill(object):
 
         # cleanup the profile if you need to
         if self.runner is not None:
-            try:
-                self.runner.profile.cleanup()
-            except OSError:
-                pass # assume profile is already cleaned up
+            self.runner.cleanup()
+
 
 class MozMillRestart(MozMill):
 
@@ -548,10 +551,13 @@ class MozMillRestart(MozMill):
         if not profile:
             profile = self.profile_class(addons=[jsbridge.extension_path, extension_path])
         self.profile = profile
-        
+
         if not runner:
-            runner = self.runner_class(profile=self.profile, 
-                                       cmdargs=["-jsbridge", str(self.jsbridge_port)])
+            runner = self.runner_class(profile=self.profile)
+
+        if not '-jsbridge' in runner.cmdargs:
+            runner.cmdargs += ['-jsbridge', '%s' % self.jsbridge_port]
+
         self.runner = runner
         self.endRunnerCalled = False
         self.add_listener(self.firePythonCallback_listener, eventType='mozmill.firePythonCallback')
@@ -578,12 +584,9 @@ class MozMillRestart(MozMill):
                                   "Components.utils.import('resource://mozmill/modules/frame.js')")
         return frame
 
-    def run_dir(self, test_dir, sleeptime=4):
+    def run_dir(self, test_dir, sleeptime=0):
         """run a directory of restart tests resetting the profile per directory"""
 
-        # Reset our Zombie counter on each directory
-        #self.zombieDetection.resetTimer()
-
         # TODO:  document this behaviour!
         if os.path.isfile(os.path.join(test_dir, 'testPre.js')):   
             pre_test = os.path.join(test_dir, 'testPre.js')
@@ -645,29 +648,45 @@ class MozMillRestart(MozMill):
         if extension_path not in profile.addons:
             profile.install_addon(extension_path)
         profile.set_preferences(profile.preferences)
+
+    def find_tests(self, tests):
+        files = []
+
+        # make sure these are all directories
+        not_dir = [ i for i in tests
+                    if not os.path.isdir(i) ]
+        if not_dir:
+            raise IOError('Restart tests must be directories (%s)' % ', '.join(not_dir))
+
+        for test_dir in tests:
+
+            # tests have to be absolute paths, for some reason
+            test_dir = os.path.abspath(test_dir)
+
+            # XXX this allows for only one sub-level of test directories
+            # is this a spec or a side-effect?
+            # If the former, it should be documented
+            test_dirs = [os.path.join(test_dir, d)
+                         for d in os.listdir(test_dir) 
+                         if d.startswith('test') and os.path.isdir(os.path.join(test_dir, d))]
+
+            if len(test_dirs):
+                files.extend(test_dirs)
+            else:
+                files.append(test_dir)
+
+        return files
     
-    def run_tests(self, test_dir, sleeptime=4):
+    def run_tests(self, tests, sleeptime=0):
 
-        self.test = test_dir
-        
-        # Zombie Counter Reset
-        #self.zombieDetector.resetTimer()
+        test_dirs = self.find_tests(tests)
+        self.tests.extend(test_dirs)
 
-        # XXX this allows for only one sub-level of test directories
-        # is this a spec or a side-effect?
-        # If the former, it should be documented
-        test_dirs = [d for d in os.listdir(os.path.abspath(os.path.expanduser(test_dir))) 
-                     if d.startswith('test') and os.path.isdir(os.path.join(test_dir, d))]
-        self.add_listener(self.firePythonCallback_listener, eventType='mozmill.firePythonCallback')
-        if not len(test_dirs):
-            test_dirs = [test_dir]
-
-        for d in test_dirs:
-            d = os.path.abspath(os.path.join(test_dir, d))
-            self.run_dir(d, sleeptime)
+        for test_dir in test_dirs:
+            self.run_dir(test_dir, sleeptime)
 
         # cleanup the profile
-        self.runner.profile.cleanup()
+        self.runner.cleanup()
 
         # Give a second for any pending callbacks to finish
         sleep(1) 
@@ -683,12 +702,7 @@ class MozMillRestart(MozMill):
         # expected that adding on more kills() for each edge case will ever
         # be able to fix a systematic issue by patching holes
         if fatal:
-            try:
-                self.runner.kill()
-                self.runner.profile.cleanup()
-            except:
-                pass
-                    
+            self.runner.cleanup()
 
 
 class CLI(jsbridge.CLI):
@@ -696,8 +710,8 @@ class CLI(jsbridge.CLI):
     module = "mozmill"
 
     parser_options = copy.copy(jsbridge.CLI.parser_options)
-    parser_options[("-t", "--test",)] = dict(dest="test", default=False, 
-                                             help="Run test file or directory.")
+    parser_options[("-t", "--test",)] = dict(dest="test", action='append', default=[],
+                                             help="Run test")
     parser_options[("-l", "--logfile",)] = dict(dest="logfile", default=None,
                                                 help="Log all events to file.")
     parser_options[("--show-errors",)] = dict(dest="showerrors", default=False, 
@@ -705,26 +719,25 @@ class CLI(jsbridge.CLI):
                                               help="Print logger errors to the console.")
     parser_options[("--report",)] = dict(dest="report", default=False,
                                          help="Report the results. Requires url to results server. Use 'stdout' for stdout.")
-    parser_options[("--showall",)] = dict(dest="showall", default=False, action="store_true",
+    parser_options[("--show-all",)] = dict(dest="showall", default=False, action="store_true",
                                          help="Show all test output.")
     parser_options[("--timeout",)] = dict(dest="timeout", type="float",
                                           default=60., 
                                           help="seconds before harness timeout if no communication is taking place")
+    parser_options[("-m", "--manifest")] = dict(dest='manifests', action='append',
+                                                help='test manifest .ini file')
+    parser_options[("--app-arg",)] = dict(dest='appArgs', action='append', default=[],
+                                          help='provides an argument to the test application')
 
     def __init__(self, *args, **kwargs):
         jsbridge.CLI.__init__(self, *args, **kwargs)
         self.mozmill = self.mozmill_class(runner_class=mozrunner.FirefoxRunner,
                                           profile_class=mozrunner.FirefoxProfile,
-                                          jsbridge_port=int(self.options.port),
+                                          jsbridge_port=self.options.port,
                                           jsbridge_timeout=self.options.timeout,
                                           )
 
-        # expand user directory and check existence for the test
-        if self.options.test:
-            self.options.test = os.path.abspath(os.path.expanduser(self.options.test))
-            if ( ( not os.path.isdir(self.options.test) )
-                 and ( not os.path.isfile(self.options.test) ) ):
-                raise Exception("Not a valid test file/directory")
+        self.tests = []
 
         # setup log formatting
         self.mozmill.add_global_listener(LoggerListener())
@@ -746,34 +759,61 @@ class CLI(jsbridge.CLI):
         return profile
 
     def run(self):
-        # XXX this is a complicated function that should probably be broken up
+
+        # read tests from manifests
+        if self.options.manifests:
+            manifest_parser = manifestparser.TestManifest(manifests=self.options.manifests)
+
+            self.tests.extend(manifest_parser.test_paths())
+
+        # expand user directory for individual tests
+        for test in self.options.test:
+            test = os.path.expanduser(test)
+            self.tests.append(test)
+                
+        # check existence for the tests
+        missing = [ test for test in self.tests
+                    if not os.path.exists(test) ]
+        if missing:
+            raise IOError("Not a valid test file/directory: %s" % ', '.join(["'%s'" % test for test in missing]))
 
         # create a Mozrunner
         runner = self.create_runner()
 
+        runner.cmdargs.extend(self.options.appArgs)
+
         # make sure the application starts in the foreground
         if '-foreground' not in runner.cmdargs:
             runner.cmdargs.append('-foreground')
-            
-        self.mozmill.start(runner=runner, profile=runner.profile)
 
-        if self.options.test:
+        try:
+            self.mozmill.start(runner=runner, profile=runner.profile)
+        except:
+            runner.cleanup()
+            raise
+
+        if self.tests:
 
             # run the tests
             disconnected = False
             try:
-                self.mozmill.run_tests(self.options.test)
+                self.mozmill.run_tests(self.tests)
             except JSBridgeDisconnectError:
                 disconnected = True
                 if not self.mozmill.userShutdownEnabled:
                     self.mozmill.report_disconnect()               
                     print 'TEST-UNEXPECTED-FAIL | Disconnect Error: Application unexpectedly closed'
+                runner.cleanup()
+            except:
+                runner.cleanup()
+                raise
+
+            # shutdown the test harness
+            self.mozmill.stop(fatal=disconnected)
 
             # print statistics and send the JSON report
             self.mozmill.report(self.options.report)
             
-            # shutdown the test harness
-            self.mozmill.stop(fatal=disconnected)
             if self.mozmill.fails or disconnected:
                 sys.exit(1)
         else:
@@ -788,19 +828,20 @@ class CLI(jsbridge.CLI):
                     runner.stop()
 
             if self.mozmill.runner is not None:
-                self.mozmill.runner.profile.cleanup()
-    
+                self.mozmill.runner.cleanup()
+
+
 class RestartCLI(CLI):
     mozmill_class = MozMillRestart
-    parser_options = copy.copy(CLI.parser_options)
-    parser_options[("-t", "--test",)] = dict(dest="test", default=False, 
-                                             help="Run test directory.")
 
 
 class ThunderbirdCLI(CLI):
     profile_class = mozrunner.ThunderbirdProfile
     runner_class = mozrunner.ThunderbirdRunner
 
+class ThunderbirdRestartCLI(RestartCLI):
+    profile_class = mozrunner.ThunderbirdProfile
+    runner_class = mozrunner.ThunderbirdRunner
 
 def enum(*sequential, **named):
     enums = dict(zip(sequential, range(len(sequential))), **named)
@@ -814,3 +855,6 @@ def tbird_cli():
 
 def restart_cli():
     RestartCLI().run()
+
+def tbird_restart_cli():
+    ThunderbirdRestartCLI().run()
